import os
import sys
import subprocess
import shlex
import pandas as pd
import requests
import pandas as pd
import numpy as np
import os
import re
import time
import traceback
import progressbar
import vcf
from ..liftover.liftover import get_liftover_positions


class DependencyBuilder:

    def __init__(self, data_dir: str = os.path.join(os.path.dirname(__file__), '..', 'data'), gnomadSV_version: str = '2.1'):
        self.data_dir = data_dir
        self.gnomadSV_version = gnomadSV_version
    

    def worker(self, cmd):
        parsed_cmd = shlex.split(cmd)
        p = subprocess.Popen(parsed_cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        p.wait()
        out, err = p.communicate()
        return out.decode() if out else err.decode()


    def get_gnomad_frequencies(self, vcf_path, output_path):

        vcf_reader = vcf.Reader(filename = vcf_path, compressed=True, encoding='ISO-8859-1')

        rows = []

        for record in vcf_reader:
            # print(record.INFO['SVTYPE'], record.FILTER)
            if record.INFO['SVTYPE'] in ('DEL','DUP') and record.FILTER == []:
                chrom = record.CHROM
                pos1 = record.POS
                pos2 = record.INFO['END']
                var_type = 'DUP' if record.INFO['SVTYPE'] == 'DUP' else 'DEL'
                freq = record.INFO['AC'][0] / record.INFO['AN']
                rows.append([chrom, pos1, pos2, var_type, freq])

        df = pd.DataFrame(rows)
        df.columns = ['chrom','start','stop','type','pop_freq']
        df.to_csv(output_path, index = False)

        
    
    def build_gnomad_db(self):
        """
        gnomadSV data is needed in order to generate population frequencies for 
        the indicated CNVs. It is quite large when unzipped (~2GB). Be sure to 
        account for this before running.
        """
        gnomad_path = os.path.join(self.data_dir, f'gnomad_v{self.gnomadSV_version}_sv.sites.vcf.gz')
        gnomad_df_path = os.path.join(self.data_dir, 'gnomad_frequencies.csv')
        if not os.path.exists(gnomad_df_path):
            url = f"https://storage.googleapis.com/gcp-public-data--gnomad/papers/2019-sv/gnomad_v{self.gnomadSV_version}_sv.sites.vcf.gz"
            cmd = f"""
                curl -s "{url}" -o {gnomad_path}
            """
            p = self.worker(cmd)

            # Parse gnomAD
            self.get_gnomad_frequencies(gnomad_path, gnomad_df_path)


    def build_breakpoints(self):
        """
        To calculate distances to centromeres and telomeres, a data frame
        containing this data must be generated. If the cytoband data for hg38 is
        not already available, it will be downloaded into the folder indicated 
        in the config file (cnv_data).
        """
        bp_path = os.path.join(self.data_dir, 'hg38_cytoband.tsv')
        if not os.path.exists(bp_path):
            url = "http://hgdownload.cse.ucsc.edu/goldenPath/hg38/database/cytoBand.txt.gz"
            cmd = f"""
                curl -s "{url}" | gunzip -c > {bp_path}
            """
            p = subprocess.Popen(cmd,  stdout=subprocess.PIPE, shell = True)
            out, err = p.communicate()
            if err is not None:
                error_message = f"""
                Failed to download the cytogeneic coordinate data at:
                {url}
                """
                raise error_message

        centromere_df_path = os.path.join(self.data_dir, 'centromeres.csv')
        if not os.path.exists(centromere_df_path):

            # Build centromere data
            centromere_df = pd.read_csv(bp_path, sep = '\t', header = None)
            centromere_df.columns = ['CHROMOSOME', 'start', 'end', 'cytoband', 'info']
            centromere_df = centromere_df.loc[centromere_df['info'] == 'acen']
            centromere_df = centromere_df.iloc[:, 0:3]
            centromere_df = centromere_df.groupby('CHROMOSOME').agg({'start': ['min'], 'end': ['max']})
            centromere_df.columns = ['start', 'end']
            centromere_df.to_csv(centromere_df_path)

        telomere_df_path = os.path.join(self.data_dir, 'telomeres.csv')
        if not os.path.exists(telomere_df_path):

            # Build telomere data
            telomere_df = pd.read_csv(bp_path, sep = '\t', header = None).iloc[:, 0:3]
            telomere_df.columns = ['CHROMOSOME', 'start', 'end']
            telomere_df = telomere_df.groupby('CHROMOSOME').agg({'start': ['min'], 'end': ['max']})
            telomere_df = pd.DataFrame(index = centromere_df.index).merge(telomere_df, how = 'left', left_index = True, right_index = True)
            telomere_df.columns = ['start', 'end']
            telomere_df.to_csv(telomere_df_path)


    def build_hi_ts_data(self):
        """
        Data downloaded from url.com. This data comes packaged with CNVpred,
        this function parses this data for use in feature generation.
        """

        hi_ts_df_path = os.path.join(self.data_dir, 'hi_ti_gene.csv')
        hi_ts_parsed_path = os.path.join(self.data_dir, 'hi_ts_regions.csv')
        if not os.path.exists(hi_ts_parsed_path):

            def parse_coordinates(row):
                pos = str(row['GRCh38'])
                pos_data = {}
                pos_data['CHROMOSOME'] = pos[:pos.find(':')]
                pos_data['START'] = pos[pos.find(':') + 1:pos.find('-')]
                pos_data['END'] = pos[pos.find('-') + 1:]
                return pos_data
            
            hi_ts_df = pd.read_csv(hi_ts_df_path, index_col=0).reset_index().rename(columns = {'index': 'type'})
            hi_ts_df['build'] = 'GRCh38'
            hi_ts_df['pos_data'] = hi_ts_df.apply(parse_coordinates, axis = 1)
            hi_ts_df = pd.concat([hi_ts_df['pos_data'].apply(pd.Series), hi_ts_df.drop(['pos_data','GRCh38'], axis = 1)], axis = 1)
            # hi_ts_df = hi_ts_df.dropna(subset = ['pLI'])
            hi_ts_df = hi_ts_df.loc[~hi_ts_df['CHROMOSOME'].isin(['', 'na'])]
            hi_ts_df = hi_ts_df.astype({'START': int, 'END': int})
            
            hi_ts_df['pos_data'] = hi_ts_df.apply(get_liftover_positions, axis = 1)
            hi_ts_df = pd.concat([hi_ts_df.drop(['pos_data','START','END'], axis = 1), hi_ts_df['pos_data'].apply(pd.Series)], axis = 1)
            hi_ts_df.to_csv(hi_ts_parsed_path, index = False)

        # Build scoring map
        hi_ts_map_path = os.path.join(self.data_dir, 'hi_ts_map.csv')
        if not os.path.exists(hi_ts_map_path):
            unique_scores = [x for x in set(hi_ts_df['HI Score'].unique()).union(set(hi_ts_df['TS Score'].unique())) if not pd.isna(x)]
            unique_scores.sort()
            hi_ts_mapped_df = pd.DataFrame({
                'HI_TS_Score': unique_scores,
                'Mapped_Score': [0, 0.25, 0.5, 1, 0, 0, 0]
            })
            hi_ts_mapped_df.to_csv(os.path.join(self.data_dir, 'hi_ts_map.csv'), index = False)


    def build_clinvar_db(self):
        """
        Download the hg38 ClinVar short variant database to find short pathogenic
        variant density
        """
        clinvar_short_path = os.path.join(self.data_dir, 'clinvar.vcf.gz')
        if not os.path.exists(clinvar_short_path):

            # Download the vcf
            cmd = f"""
            wget https://ftp.ncbi.nlm.nih.gov/pub/clinvar/vcf_GRCh38/clinvar.vcf.gz -P {self.data_dir}
            """
            p = self.worker(cmd)

            # Create an index file
            # cmd = f"""
            # {config['tabix']} -f {clinvar_short_path}
            # """
            cmd = f"""
            tabix -f {clinvar_short_path}
            """
            p = self.worker(cmd)


    def get_mim2gene(self):
        """
        Get the mim2gene data
        """
        mim2_gene_url = 'https://ftp.ncbi.nih.gov/gene/DATA/mim2gene_medgen'
        output_file = os.path.join(self.data_dir, 'mim2gene_medgen')
        cmd = f"""
                curl -s "{mim2_gene_url}" -o {output_file}
            """
        p = self.worker(cmd)


    def build_all(self):
        """
        Build all dependencies and update a progress bar as we chug along
        """

        # Intialize progress bar
        bar_widgets = [progressbar.FormatLabel('Downloading and parsing gnomAD'), ' ', progressbar.Bar('=', '[', ']'), ' ', progressbar.Percentage(), ' ', progressbar.Timer()]
        bar = progressbar.ProgressBar(maxval=5, \
            widgets=bar_widgets)
        bar.start()

        # Download and unpack gnomAD data
        self.build_gnomad_db()
        bar.update(1)
        bar_widgets[0] = progressbar.FormatLabel('Downloading chromosomal breakpoints')
        
        # Get chromosomal breakpoint data
        self.build_breakpoints()
        bar.update(2)
        bar_widgets[0] = progressbar.FormatLabel('Downloading HI/TS data')

        # Download and unpack haploinsufficiency (HI) and triplosensitivity (TS) data
        self.build_hi_ts_data()
        bar.update(3)
        bar_widgets[0] = progressbar.FormatLabel('Downloading ClinVar short variants')

        # Download and unpack ClinVar short variants
        self.build_clinvar_db()
        bar.update(4)
        bar_widgets[0] = progressbar.FormatLabel('Downloading mim2gene file')

        # Download OMIM to gene annotations
        self.get_mim2gene()
        bar.update(5)
        bar_widgets[0] = progressbar.FormatLabel('Dependencies ready')

        bar.finish()



from .feature_builder.feature_builder import FeatureBuilder
from .dependency_builder.dependency_builder import DependencyBuilder
from .classifier.classifier import Classifier

# Set them as attributes of the package
__all__ = ['FeatureBuilder', 'DependencyBuilder', 'Classifier']



import pandas as pd
import os
import re
import math
import json
import subprocess
import sys
import progressbar
import vcf
from sklearn.preprocessing import MinMaxScaler
from ..liftover.liftover import get_liftover_positions


class FeatureBuilder:

    def __init__(self, variant_df: pd.DataFrame, data_dir: str, ref: str):

        # Get root path
        self.data_dir = data_dir
        self.ref = ref
        self.centromere_df = pd.read_csv(os.path.join(self.data_dir, 'centromeres.csv'), index_col = 0)
        self.telomere_df = pd.read_csv(os.path.join(self.data_dir, 'telomeres.csv'),  index_col = 0)

        # Replace nulls with -999 for genes with no gene associations
        self.mim2gene = pd.read_csv(os.path.join(self.data_dir, 'mim2gene_medgen'), sep = '\t')
        self.mim2gene.loc[self.mim2gene['GeneID'].str.contains('-'), 'GeneID'] = -999
        self.mim2gene = self.mim2gene.astype({'GeneID': int})
        self.mim2gene.columns = ['OMIM ID'] + self.mim2gene.columns[1:].to_list()

        self.gnomadSV = pd.read_csv(os.path.join(self.data_dir, 'gnomad_frequencies.csv'), dtype = {'chrom': str})
        self.hi_ts_df = pd.read_csv(os.path.join(self.data_dir, 'hi_ts_regions.csv'))
        self.hi_ts_mapped_df = pd.read_csv(os.path.join(self.data_dir, 'hi_ts_map.csv'))
        self.variant_df = variant_df

        # Create a breakpoint dataframe to determine chromosome arm length
        self.breakpoint_df = self.centromere_df.merge(self.telomere_df, on = 'CHROMOSOME', suffixes = ('_cen','_tel'))
        self.breakpoint_df['arm1_len'] = self.breakpoint_df['start_cen'] -  self.breakpoint_df['start_tel']
        self.breakpoint_df['arm2_len'] = self.breakpoint_df['end_tel'] - self.breakpoint_df['end_cen']


    def get_features(self):#, clinvar_ids: list): 
        """
        Build feature matrix that is compatible with sklearn / tensorflow for model 
        training. Features include:
            1. Number of genes spanned when gene is at least half spanned, partial when 
            less than half of gene is spanned
            2. Population frequency of CNV
            3. Copy Number
            4. Genome position
            5. Normalized distance to centromere
            6. Normalized distance to telemere
            7. CG content
            8. Haploinsufficient genomic region overlap (or triplosensitive)
            9. Length of CNV (in bp)
            10. GO terms (some sort of intersect function)
            11. Average dbNSFP
        """

        def get_bp_length(row):
            return row['END'] - row['START'] + 1
        
        def get_all_genes(row, gene_position_df):
            """
            Query the gene position table to find the intersecting genes
            """

            intersect_df = gene_position_df[
                (row['CHROMOSOME'] == gene_position_df['Chromosomes'])
                    &
                (
                    (
                        (row['START'] <= gene_position_df['GRCh38_start'])
                            &
                        (row['END'] >= gene_position_df['GRCh38_start'])
                    )
                        |
                    (
                        (row['START'] <= gene_position_df['GRCh38_stop'])
                            &
                        (row['END'] >= gene_position_df['GRCh38_stop'])
                    )
                        |
                    (
                        (row['START'] >= gene_position_df['GRCh38_start'])
                            &
                        (row['END'] <= gene_position_df['GRCh38_stop'])
                    )
                )
            ]
                
            return intersect_df['GeneID'].to_list()


        def get_omim_disease_genes(row):
            """
            Using the gene-disease OMIM table, find a list of genes that are 
            associated with a rare disease. Return the length of the resulting 
            gene list.
            """
            associated_diseases = self.mim2gene.loc[(self.mim2gene['GeneID'].isin(row["gene_info"])) & (self.mim2gene['type'] == 'phenotype')]
            return len(associated_diseases.index)


        def get_gene_count(row):
            """
            Count the total number of genes
            """
            return len(row["gene_info"])


        def get_centromere_distance(row):
            """
            Use the previously downloaded centromere coordinate data to obtain 
            the distance of the CNV to the centromere (normalized by chromosome 
            arm length)
            """
            cen_start, cen_end = self.breakpoint_df.loc[f"chr{row['CHROMOSOME']}", 'start_cen'], self.breakpoint_df.loc[f"chr{row['CHROMOSOME']}", 'end_cen']
            arm1_len, arm2_len = self.breakpoint_df.loc[f"chr{row['CHROMOSOME']}", 'arm1_len'], self.breakpoint_df.loc[f"chr{row['CHROMOSOME']}", 'arm2_len']
            if row['END'] < cen_start:
                return (cen_start - row['END']) / arm1_len
            else:
                return (row['START'] - cen_end) / arm2_len


        def get_telomere_distance(row):
            """
            Use the previously downloaded telomere coordinate data to obtain the 
            distance of the CNV to the telomere (normalized by chromosome arm 
            length)
            """
            tel_start, tel_end = self.breakpoint_df.loc[f"chr{row['CHROMOSOME']}", 'start_tel'], self.breakpoint_df.loc[f"chr{row['CHROMOSOME']}", 'end_tel']
            cen_start, cen_end = self.breakpoint_df.loc[f"chr{row['CHROMOSOME']}", 'start_cen'], self.breakpoint_df.loc[f"chr{row['CHROMOSOME']}", 'end_cen']
            arm1_len, arm2_len = self.breakpoint_df.loc[f"chr{row['CHROMOSOME']}", 'arm1_len'], self.breakpoint_df.loc[f"chr{row['CHROMOSOME']}", 'arm2_len']
            if row['END'] < cen_start:
                return (row['START'] - tel_start) / arm1_len
            else:
                return (tel_end - row['END']) / arm2_len


        def get_gc_content(row):
            command = f"samtools faidx {self.ref} chr{row['CHROMOSOME']}:{row['START']}-{row['END']}"
            p = subprocess.Popen(command,  stdout=subprocess.PIPE, shell = True)
            out, err = p.communicate()
            s = "".join(out.decode().strip().split()[1:])
            s = s.replace('N','') # Don't count N values
            cg_len = len(s.replace('A','').replace('T',''))
            try:
                return round(cg_len/len(s), 3)
            except:
                print(s)

        def get_hi_ts_regions(row):
            """
            Haploinsucciency(HS) and triplosensitive(TS) regions have been 
            downloaded by the dependency builder. This function reads the 
            donwloaded data set and determines if the CNV provided in "row"
            intersects with any HS or TS regions, and to what extent.
            """

            def get_region_coverage(row, cnv_start, cnv_end):
                left = max(cnv_start, row['START'])
                right = min(cnv_end, row['END'])
                num = left - right
                denom = row['START'] - row['END']
                return num / denom

            def get_hi_ts_scores(row):
                data = {}
                data['HI_SCORE'] = self.hi_ts_mapped_df.loc[self.hi_ts_mapped_df['HI_TS_Score'] == row['HI Score']].reset_index().loc[0, 'Mapped_Score']
                data['TS_SCORE'] = self.hi_ts_mapped_df.loc[self.hi_ts_mapped_df['HI_TS_Score'] == row['TS Score']].reset_index().loc[0, 'Mapped_Score']
                return data

            # Filter to only include regions intersecting with CNV
            df = self.hi_ts_df.loc[
                (self.hi_ts_df['CHROMOSOME'] == row['CHROMOSOME']) 
                        & 
                    (
                        (self.hi_ts_df['START'].between(row['START'], row['END']))
                            |
                        (self.hi_ts_df['START'].between(row['START'], row['END']))
                    )
                ]

            # Only keep the gene entries
            # print(row['clinvar_id'], row['gene_count'], f'chr{row["chrom"]}', row['GRCh38_start'], row['GRCh38_end'], len(df.index), len(df[df['type'] != 'R'].index), len(df[(df['%HI'] != '‐') | (df['pLI'] != '‐') | (df['LOEUF'] != '‐')].index), row['gene_info'])
            df = df[(df['%HI'] != '‐') & (df['pLI'] != '‐') & (df['LOEUF'] != '‐')]

            # Exit if the dataframe is empty
            if df.empty:
                return {
                    'HI_SCORE': 0,
                    'TS_SCORE': 0,
                    '%HI': 100,
                    'pLI': 0,
                    'LOEUF': 2 # max value in HI table
                    }

            # Get coverage percentage for each intersecting region
            df['percent_coverage'] = df.apply(
                get_region_coverage, 
                cnv_start = row['START'], 
                cnv_end = row['END'],
                axis = 1)

            # Map HI and TS scores from pre-defined values
            df['data'] = df.apply(get_hi_ts_scores, axis = 1)
            df = pd.concat([df.drop(['HI Score', 'TS Score', 'data'], axis = 1), df['data'].apply(pd.Series)], axis = 1)

            # Calculate final overlap * (TS/HS)
            df['adjusted_HI_score'] = df['percent_coverage'] * df['HI_SCORE']
            df['adjusted_TS_score'] = df['percent_coverage'] * df['TS_SCORE']

            # Fill null values in float columns
            for col in ['%HI', 'pLI', 'LOEUF']:
                try:
                    df[col] = df[col].astype(float)
                except Exception as e:
                    print(type(e), e)
                    print(df)

            df['adjusted_%HI'] = df['percent_coverage'] * df['%HI']
            df['adjusted_pLI'] = df['percent_coverage'] * df['pLI']
            df['adjusted_LOEUF'] = df['percent_coverage'] * df['LOEUF']

            # Return the score of the worst intersecting HI/TS region
            return {
                'HI_SCORE': df['adjusted_HI_score'].max(),
                'TS_SCORE': df['adjusted_TS_score'].max(),
                '%HI': df['adjusted_%HI'].min(),
                'pLI': df['adjusted_pLI'].max(),
                'LOEUF': df['adjusted_LOEUF'].min()
                }


        def get_population_frequency(row):
            """
            Gathers all gnomAD SV records that intersect with the CNV provided 
            in "row". Returns the average of the population frequencies observed 
            in intersecting variants
            """

            def get_region_coverage(row, cnv_start, cnv_end):
                left = max(cnv_start, row['start'])
                right = min(cnv_end, row['stop'])
                num = left - right
                denom = row['start'] - row['stop']
                return num / denom

            gnomad_df = self.gnomadSV.loc[
                    (self.gnomadSV['start'].between(row['START'], row['END'])) 
                        | 
                    (self.gnomadSV['stop'].between(row['START'], row['END']))
                ]

            if not gnomad_df.empty:
                gnomad_df['percent_coverage'] = gnomad_df.apply(
                    get_region_coverage, 
                    cnv_start = row['START'], 
                    cnv_end = row['END'],
                    axis = 1)

            else:
                gnomad_df['percent_coverage'] = 0

            # Only consider variants that cover at least half of the CNV
            gnomad_df = gnomad_df.loc[gnomad_df['percent_coverage'] >= 0.5]

            return gnomad_df['pop_freq'].mean()

        
        def get_clinvar_path_density(row, clinvar_reader):
            path_count = 0
            query = clinvar_reader.fetch(chrom = row['CHROMOSOME'], start = row['START'], end = row['END'])
            for q in query:
                try:
                    if re.search('PATHOGENIC', q.INFO['CLNSIG'][0].upper()):
                        path_count += 1
                except:
                    pass
            return path_count

        def get_ptriplo(row, c_data):
            try:
                max_ = c_data[c_data['geneID'].isin(row['gene_info'])]['pHaplo'].max()
                return max_ if not pd.isna(max_) else 0
            except:
                return 0
        
        def get_htriplo(row, c_data):
            try:
                max_ = c_data[c_data['geneID'].isin(row['gene_info'])]['pTriplo'].max()
                return max_ if not pd.isna(max_) else 0
            except:
                return 0

        def get_collins_data(row, c_data_full, col, func_):
            try:
                if func_ == 'max':
                    max_ = c_data_full[c_data_full['geneID'].isin(row['gene_info'])][col].max()
                    return max_ if not pd.isna(max_) else 0
                elif func_ == 'min':
                    min_ = c_data_full[c_data_full['geneID'].isin(row['gene_info'])][col].min()
                    return min_ if not pd.isna(min_) else 1
                elif func_ == 'mean':
                    mean_ = c_data_full[c_data_full['geneID'].isin(row['gene_info'])][col].mean()
                    return mean_ if not pd.isna(mean_) else 0.5
            except:
                if func_ == 'max':
                    return 0
                elif func_ == 'min':
                    return 1
                else:
                    return 0

        cnv_df = self.variant_df.copy()

        # Load dependencies
        gene_position_df = pd.read_csv(os.path.join(self.data_dir, 'gene_positions.csv'))

        # Intialize progress bar
        bar_widgets = [progressbar.FormatLabel('Building Features'), ' ', progressbar.Bar('=', '[', ']'), ' ', progressbar.Percentage(), ' ', progressbar.Timer()]
        bar = progressbar.ProgressBar(maxval=14, \
            widgets=bar_widgets)
        bar.start()
        bar_widgets[0] = progressbar.FormatLabel('Ensuring positions are in hg38')

        # Map to hg38
        # cnv_df['build'] = config['build']
        cnv_df['build'] = 'GRCh38'
        cnv_df['variant'] = cnv_df.apply(get_liftover_positions, axis = 1)
        cnv_df = pd.concat([cnv_df.drop(['variant','START','END'], axis = 1), cnv_df['variant'].apply(pd.Series)], axis = 1)
        cnv_df = cnv_df.astype({'START': 'int32', 'END': 'int32'})
        original_cols = list(cnv_df.columns)
        # cnv_df = cnv_df[['CHROMOSOME','START','END','CHANGE']]

        # Get features
        bar.update(1)
        bar_widgets[0] = progressbar.FormatLabel('Collecting CNV length')
        cnv_df['BP_LEN'] = cnv_df.apply(get_bp_length, axis = 1)

        bar.update(2)
        bar_widgets[0] = progressbar.FormatLabel('Collecting gene data')
        cnv_df['gene_info'] = cnv_df.apply(get_all_genes, gene_position_df = gene_position_df, axis = 1)

        bar.update(3)
        bar_widgets[0] = progressbar.FormatLabel('Calculating gene count')
        cnv_df['GENE_COUNT'] = cnv_df.apply(get_gene_count, axis = 1)

        bar.update(4)
        bar_widgets[0] = progressbar.FormatLabel('Calculating OMIM disease count')
        cnv_df['DISEASE_COUNT'] = cnv_df.apply(get_omim_disease_genes, axis = 1)

        bar.update(5)
        bar_widgets[0] = progressbar.FormatLabel('Calculating centromere distance count')
        cnv_df['CENT_DIST'] = cnv_df.apply(get_centromere_distance, axis = 1)

        bar.update(6)
        bar_widgets[0] = progressbar.FormatLabel('Calculating telomere distance count')
        cnv_df['TEL_DIST'] = cnv_df.apply(get_telomere_distance, axis = 1)

        bar.update(7)
        bar_widgets[0] = progressbar.FormatLabel('Calculating HI/TS coverage')
        cnv_df['hi_ts_region_scores'] = cnv_df.apply(get_hi_ts_regions, axis = 1)
        cnv_df = pd.concat([cnv_df.drop('hi_ts_region_scores', axis = 1), cnv_df['hi_ts_region_scores'].apply(pd.Series) ], axis = 1)

        bar.update(8)
        bar_widgets[0] = progressbar.FormatLabel('Calculating GC content')
        cnv_df['GC_CONTENT'] = cnv_df.apply(get_gc_content, axis = 1)

        bar.update(9)
        bar_widgets[0] = progressbar.FormatLabel('Collecting population frequencies')
        cnv_df['POP_FREQ'] = cnv_df.apply(get_population_frequency, axis = 1).fillna(0)

        bar.update(10)
        bar_widgets[0] = progressbar.FormatLabel('Collecting ClinVar short variant density')

        # Create reader of ClinVar short variants
        clinvar_path = os.path.join(self.data_dir, 'clinvar.vcf.gz')
        clinvar_reader = vcf.Reader(filename = clinvar_path, compressed=True, encoding='ISO-8859-1')
        cnv_df['CLINVAR_DENSITY'] = cnv_df.apply(get_clinvar_path_density, clinvar_reader = clinvar_reader, axis = 1)
        
        # bar.update(11)
        # bar_widgets[0] = progressbar.FormatLabel('ClinVar density generated')

        # # Get maximum haploinsuffeciency and triplosensitivity probabilities 
        # collins_path = os.path.join(self.data_dir, 'TS_score_geneid.csv')
        # collins_df = pd.read_csv(collins_path)
        # cnv_df['P_TRIPLO'] = cnv_df.apply(get_ptriplo, c_data = collins_df, axis = 1)
        # cnv_df['P_HAPLO'] = cnv_df.apply(get_htriplo, c_data = collins_df, axis = 1)

        # bar.update(12)
        # bar_widgets[0] = progressbar.FormatLabel('Probability HI/TS generated')

        # # Get other collins features
        # collins_full_path = os.path.join(self.data_dir, 'collins_raw_geneid.csv')
        # collins_full_df = pd.read_csv(collins_full_path)
        # cnv_df['gnomad_oe_lof'] = cnv_df.apply(get_collins_data, c_data_full = collins_full_df, col = 'gnomad_oe_lof', func_ = 'min', axis = 1)
        # cnv_df['gnomad_oe_lof_upper'] = cnv_df.apply(get_collins_data, c_data_full = collins_full_df, col = 'gnomad_oe_lof_upper', func_ = 'min', axis = 1)
        # cnv_df['gnomad_pLI'] = cnv_df.apply(get_collins_data, c_data_full = collins_full_df, col = 'gnomad_pLI', func_ = 'max', axis = 1)
        # cnv_df['gnomad_lof_z'] = cnv_df.apply(get_collins_data, c_data_full = collins_full_df, col = 'gnomad_lof_z', func_ = 'max', axis = 1)
        # cnv_df['gnomad_mis_z'] = cnv_df.apply(get_collins_data, c_data_full = collins_full_df, col = 'gnomad_mis_z', func_ = 'max', axis = 1)
        # cnv_df['gnomad_oe_mis_upper'] = cnv_df.apply(get_collins_data, c_data_full = collins_full_df, col = 'gnomad_oe_mis_upper', func_ = 'min', axis = 1)
        # cnv_df['episcore'] = cnv_df.apply(get_collins_data, c_data_full = collins_full_df, col = 'episcore', func_ = 'max', axis = 1)
        # cnv_df['sHet'] = cnv_df.apply(get_collins_data, c_data_full = collins_full_df, col = 'sHet', func_ = 'max', axis = 1)
        # cnv_df['exon_phastcons'] = cnv_df.apply(get_collins_data, c_data_full = collins_full_df, col = 'exon_phastcons', func_ = 'max', axis = 1)
        # cnv_df['hurles_hi'] = cnv_df.apply(get_collins_data, c_data_full = collins_full_df, col = 'hurles_hi', func_ = 'max', axis = 1)
        # cnv_df['rvis'] = cnv_df.apply(get_collins_data, c_data_full = collins_full_df, col = 'rvis', func_ = 'max', axis = 1)
        # cnv_df['promoter_cpg_count'] = cnv_df.apply(get_collins_data, c_data_full = collins_full_df, col = 'promoter_cpg_count', func_ = 'max', axis = 1)
        # cnv_df['eds'] = cnv_df.apply(get_collins_data, c_data_full = collins_full_df, col = 'eds', func_ = 'max', axis = 1)
        # cnv_df['promoter_phastcons'] = cnv_df.apply(get_collins_data, c_data_full = collins_full_df, col = 'promoter_phastcons', func_ = 'max', axis = 1)
        # cnv_df['cds_length'] = cnv_df.apply(get_collins_data, c_data_full = collins_full_df, col = 'cds_length', func_ = 'max', axis = 1)
        # cnv_df['gnomad_pNull'] = cnv_df.apply(get_collins_data, c_data_full = collins_full_df, col = 'gnomad_pNull', func_ = 'min', axis = 1)
        


        bar.update(11)
        # bar.update(13)
        bar_widgets[0] = progressbar.FormatLabel('All features generated, normalizing')


        # Fill NAs
        cnv_df = cnv_df.fillna({
            'GC_CONTENT': .5,
            'CLINVAR_DENSITY': 0,
            'POP_FREQ': 0
        })


        # Normalize features
        cnv_norm_df = cnv_df.copy()
        scaler = MinMaxScaler()

        cols_to_normalize = [
            'BP_LEN','GENE_COUNT','DISEASE_COUNT','CENT_DIST','TEL_DIST', 
            'HI_SCORE','TS_SCORE','%HI','pLI','LOEUF','GC_CONTENT', 
            'POP_FREQ','CLINVAR_DENSITY'
        ]
        # cols_to_normalize = [
        #     'BP_LEN','GENE_COUNT','DISEASE_COUNT','CENT_DIST','TEL_DIST', 
        #     'HI_SCORE','TS_SCORE','%HI','pLI','LOEUF','GC_CONTENT', 
        #     'POP_FREQ','CLINVAR_DENSITY','P_TRIPLO','P_HAPLO','gnomad_oe_lof',
        #     'gnomad_oe_lof_upper','gnomad_pLI','gnomad_lof_z','gnomad_mis_z',
        #     'gnomad_oe_mis_upper','episcore','sHet','exon_phastcons','hurles_hi',
        #     'rvis','promoter_cpg_count','eds','promoter_phastcons','cds_length',
        #     'gnomad_pNull'
        # ]

        # for col in cols_to_normalize:
        #     cnv_norm_df[col] = scaler.fit_transform(cnv_norm_df[col].values.reshape(-1,1))

        bar_widgets[0] = progressbar.FormatLabel('Feature generation complete')
        bar.update(12)
        # bar.update(14)
        bar.finish()

        cols = ['CHROMOSOME','START','END','CHANGE'] 
        cols = cols + [oc for oc in original_cols if oc not in cols] + ['gene_info'] + cols_to_normalize
        cols = [c for c in cols if c != 'build']

        # self.norm_feature_df, self.feature_df = cnv_norm_df[cols], cnv_df[cols]
        self.feature_df = cnv_df[cols]
        



from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="CNVoyant",
    version="0.1.9",
    author="Rob Schuetz",
    author_email="robert.schuetz@nationwidechildrens.org",
    description="Copy Number Variant Pathogenicity Classifier",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Rob-Schuetz/CNVoyant", 
    project_urls={
        "Bug Tracker": "https://github.com/Rob-Schuetz/CNVoyant/issues",
    },
    license="MIT",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    packages=find_packages(),
    install_requires=[
        'numpy==1.23.4',
        'onnxruntime==1.12.1',
        'pandas==1.5.0',
        'progressbar==2.5',
        'requests==2.28.1',
        'setuptools==65.6.3',
        'xgboost==1.7.1',
        'pyvcf==0.6.8',
        'scikit-learn==1.0.1',
        'pickleshare==0.7.5',
        'uuid==1.30',
        'pysam==0.22.0'
    ],
    python_requires=">=3.6",
    include_package_data=True
)



import os
import uuid
import shlex
import subprocess
import json
import pandas as pd


def worker(cmd):
    parsed_cmd = shlex.split(cmd)
    p = subprocess.Popen(parsed_cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    p.wait()
    out, err = p.communicate()
    return out.decode() if out else err.decode()


def get_liftover_data():
    with open(os.path.join(os.path.dirname(__file__), 'paths.json'),'r') as f:
        liftover_data = json.load(f)
    liftover_data = {k:v if v.startswith('/') else os.path.join(os.path.dirname(__file__), v) for k,v in liftover_data.items()}
    return liftover_data


def get_liftover_positions(row):
    """
    Get the GRCh38 positions for every CNV. Requires a pointer to a
    previously downloaded liftover executable. Can be downloaded with the 
    DependencyBuilder object.
    """

    data = get_liftover_data()
    
    coordinates = {}
    liftover_executable_path = data['liftover_executable_path']
    input_filename = os.path.join(data['liftover_temp_dir'], str(uuid.uuid4()))
    lifted_output_filename = os.path.join(data['liftover_temp_dir'], str(uuid.uuid4()))
    unlifted_output_filename = os.path.join(data['liftover_temp_dir'], str(uuid.uuid4()))
            
    if row['build'] == 'GRCh38':
        return {'START': row['START'], 'END': row['END']}

    # Write coordinates to a temporary file
    pd.DataFrame({
        'chrom': ['chr' + str(row['CHROMOSOME'])],
        'pos': [row['START']],
        'pos1': [row['END']]
    }).to_csv(input_filename, sep = '\t', index = False, header = False)

    # Pass to the liftover command line executable
    chain_file = data['liftover_hg19tohg38_chain_path']
    cmd = f'{liftover_executable_path} {input_filename} {chain_file} {lifted_output_filename} {unlifted_output_filename}'
    p = worker(cmd)

    # Read in the results
    try:
        lifted_coordinates = pd.read_csv(lifted_output_filename, sep = '\t', header = None)
    except:
        print(f"Failed to liftover at {'chr' + row['CHROMOSOME']}:{row['START']}-{row['END']}")
        lifted_coordinates = pd.read_csv(unlifted_output_filename, sep = '\t', header = None, comment='#')
    coordinates['START'] = lifted_coordinates.iloc[0, 1]
    coordinates['END'] = lifted_coordinates.iloc[0, 2]
    
    # Remove temp files
    for f in [input_filename, lifted_output_filename, unlifted_output_filename]:
        os.remove(f)
    
    return coordinates



import sys
import os
import pickle
import pandas as pd
import numpy as np
from sklearn import ensemble
import xgboost
from sklearn.pipeline import Pipeline
import onnxruntime as rt
import progressbar


class Classifier():

    def __init__(self, norm = True):
        self.random_seed = 42
        self.root_path = os.path.dirname(__file__)
        self.onnx = True
        self.dup_model = None
        self.del_model = None
        self.norm = norm

    def train(self, input: pd.DataFrame):
        """
        Generate duplication and deletion models based on previously discovered
        optimal hyper-parameters.
        Input: 
            input: pandas DatFrame generate from FeatureBuilder
            label: column title indicating the label of the CNVs
        Output:
            tuple containing a duplication and deletion model
        """

        # Intialize progress bar
        bar_widgets = [progressbar.FormatLabel('Formatting training data'), ' ', progressbar.Bar('=', '[', ']'), ' ', progressbar.Percentage(), ' ', progressbar.Timer()]
        bar = progressbar.ProgressBar(maxval=3, \
            widgets=bar_widgets)
        bar.start()

        # Limit to feature, label, and CNV type columns
        # feature_cols = [
        #     'BP_LEN','GENE_COUNT','DISEASE_COUNT','CENT_DIST','TEL_DIST', 
        #     'HI_SCORE','TS_SCORE','%HI','pLI','LOEUF','GC_CONTENT', 
        #     'POP_FREQ','CLINVAR_DENSITY','P_TRIPLO','P_HAPLO','gnomad_oe_lof',
        #     'gnomad_oe_lof_upper','gnomad_pLI','gnomad_lof_z','gnomad_mis_z',
        #     'gnomad_oe_mis_upper','episcore','sHet','exon_phastcons','hurles_hi',
        #     'rvis','promoter_cpg_count','eds','promoter_phastcons','cds_length',
        #     'gnomad_pNull'
        # ]
        feature_cols = [
            'BP_LEN','GENE_COUNT','DISEASE_COUNT','CENT_DIST','TEL_DIST', 
            'HI_SCORE','TS_SCORE','%HI','pLI','LOEUF','GC_CONTENT', 
            'POP_FREQ','CLINVAR_DENSITY'
        ]
        label = 'LABEL'

        # Divide training data by CNV type
        XY = input.copy()

        if self.norm:
            with open(os.path.join(self.root_path, 'models', 'scalers.pickle'), 'rb') as f:
                col_scalers = pickle.load(f)
            for col in feature_cols:
                XY[col] = col_scalers[col].transform(XY[col].values.reshape(-1,1))

        XY_dup = XY[XY['CHANGE'] == 'DUP']
        XY_del = XY[XY['CHANGE'] == 'DEL']

        X_dup = XY_dup[feature_cols]
        X_del = XY_del[feature_cols]
        Y_dup = XY_dup[label]
        Y_del = XY_del[label]

        X_dup.columns = [f'f{i}' for i in range(len(X_dup.columns))]
        X_del.columns = [f'f{i}' for i in range(len(X_del.columns))]

        # if len(X.index) > 40000:
        #     X.to_csv('/igm/projects/CNVpred/notebooks/training_data_full.csv')
        # else:
        #     X.to_csv('/igm/projects/CNVpred/notebooks/training_data_limited.csv')

        # Map Y to label codes
        def get_map(row, map):
            return map[row]
        
        map = {
            'Benign': 0,
            'VUS': 1,
            'Pathogenic': 2
        }

        Y_del = np.array(Y_del.apply(get_map, map = map))
        Y_dup = np.array(Y_dup.apply(get_map, map = map))

        # Define optimal hyperparameters
        rf_del = {
            'n_estimators': 150, 'min_samples_split': 5, 'min_samples_leaf': 3, 
            'max_features': 'sqrt', 'max_depth': None, 'bootstrap': False
            }

        xgb_dup = {
            'tree_method': 'approx', 'reg_lambda': 1, 'reg_alpha': 0,
            'n_estimators': 100, 'max_depth': 10, 'gamma': 0, 'eta': 0.1, 
            'booster': 'gbtree', 'base_score': 0.5
         }
        
        # Generate models
        bar.update(1)
        bar_widgets[0] = progressbar.FormatLabel('Training deletion model')
        self.del_model = ensemble.RandomForestClassifier(
            **rf_del,
            random_state = self.random_seed
        ).fit(X_del, Y_del)
        bar.update(2)
        bar_widgets[0] = progressbar.FormatLabel('Training duplication model')

        self.dup_model = xgboost.XGBClassifier(
            **xgb_dup,
            random_state = self.random_seed
        ).fit(X_dup, Y_dup)
        bar.update(3)
        bar_widgets[0] = progressbar.FormatLabel('Model training complete')

        self.pipe = Pipeline([('xgb', xgboost.XGBClassifier(
            **xgb_dup,
            random_state = self.random_seed
        ))]).fit(X_dup, Y_dup)

        # with open('/igm/home/rsrxs003/rnb/output/BL-226/pickled_models/dup_pipe.pickle', 'wb') as f:
        #     pickle.dump(pipe, f)

        # with open('/igm/home/rsrxs003/rnb/output/BL-226/pickled_models/del.pickle', 'wb') as f:
        #     pickle.dump(self.del_model, f)
        
        # with open('/igm/home/rsrxs003/rnb/output/BL-226/pickled_models/dup.pickle', 'wb') as f:
        #     pickle.dump(self.dup_model, f)

        self.onnx = False


    def predict(self, input: pd.DataFrame):
        """
        Utilize either the default or re-trained models to return a pandas
        DataFrame containing prediction probabilities for pathogenic, VUS, or
        benign clinical classification.
            Input: pandas DataFrame generated from FeatureBuilder
        """

        # Intialize progress bar
        bar_widgets = [progressbar.FormatLabel('Formatting input data'), ' ', progressbar.Bar('=', '[', ']'), ' ', progressbar.Percentage(), ' ', progressbar.Timer()]
        bar = progressbar.ProgressBar(maxval=4, \
            widgets=bar_widgets)
        bar.start()

        # Limit to feature columns
        # feature_cols = [
        #     'BP_LEN','GENE_COUNT','DISEASE_COUNT','CENT_DIST','TEL_DIST', 
        #     'HI_SCORE','TS_SCORE','%HI','pLI','LOEUF','GC_CONTENT', 
        #     'POP_FREQ','CLINVAR_DENSITY','P_TRIPLO','P_HAPLO','gnomad_oe_lof',
        #     'gnomad_oe_lof_upper','gnomad_pLI','gnomad_lof_z','gnomad_mis_z',
        #     'gnomad_oe_mis_upper','episcore','sHet','exon_phastcons','hurles_hi',
        #     'rvis','promoter_cpg_count','eds','promoter_phastcons','cds_length',
        #     'gnomad_pNull'
        # ]

        feature_cols = [
            'BP_LEN','GENE_COUNT','DISEASE_COUNT','CENT_DIST','TEL_DIST', 
            'HI_SCORE','TS_SCORE','%HI','pLI','LOEUF','GC_CONTENT', 
            'POP_FREQ','CLINVAR_DENSITY'
        ]


        # Divide training data by CNV type
        XY = input.copy()

        if self.norm:
            with open(os.path.join(self.root_path, 'models', 'scalers.pickle'), 'rb') as f:
                col_scalers = pickle.load(f)
            for col in feature_cols:
                XY[col] = col_scalers[col].transform(XY[col].values.reshape(-1,1))

        XY_dup = XY[XY['CHANGE'] == 'DUP']
        XY_del = XY[XY['CHANGE'] == 'DEL']

        X_dup = XY_dup[feature_cols]
        X_del = XY_del[feature_cols]

        X_dup.columns = [f'f{i}' for i in range(len(X_dup.columns))]
        X_del.columns = [f'f{i}' for i in range(len(X_del.columns))]


        # Generate predictions
        bar_widgets[0] = progressbar.FormatLabel('Obtaining DEL predictions')
        bar.update(1)

        if self.onnx:
            
            sess = rt.InferenceSession(os.path.join(self.root_path, 'models', "del_norm.onnx"))
            input_name = sess.get_inputs()[0].name
            del_pred = sess.run(None, {input_name: np.array(X_del, dtype = 'f')})
            del_pred = pd.DataFrame(del_pred[1]).rename(columns = {0:'BENIGN',1:'VUS',2:'PATHOGENIC'})
            
            bar_widgets[0] = progressbar.FormatLabel('Obtaining DUP predictions')
            bar.update(2)

            sess = rt.InferenceSession(os.path.join(self.root_path, 'models', "dup_norm.onnx"))
            input_name = sess.get_inputs()[0].name
            dup_pred = sess.run(None, {input_name: np.array(X_dup, dtype = 'f')})
            dup_pred = pd.DataFrame(dup_pred[1]).rename(columns = {0:'BENIGN',1:'VUS',2:'PATHOGENIC'})

        else:
            del_pred = self.del_model.predict_proba(X_del)
            del_pred = pd.DataFrame(del_pred).rename(columns = {0:'BENIGN',1:'VUS',2:'PATHOGENIC'})

            bar_widgets[0] = progressbar.FormatLabel('Obtaining DUP predictions')
            bar.update(2)

            dup_pred = self.dup_model.predict_proba(X_dup)
            dup_pred = pd.DataFrame(dup_pred).rename(columns = {0:'BENIGN',1:'VUS',2:'PATHOGENIC'})

        bar_widgets[0] = progressbar.FormatLabel('Reformatting and saving predictions')
        bar.update(3)
        
        # Concatenate positions and predictions
        XY_dup_orig = XY_dup[[ c for c in XY_dup.columns if c not in feature_cols + list(dup_pred.columns) ]]
        XY_del_orig = XY_del[[ c for c in XY_del.columns if c not in feature_cols + list(dup_pred.columns) ]]

        XY_dup_orig = XY_dup_orig.reset_index(drop=True)
        X_dup = X_dup.reset_index(drop=True)
        X_dup.columns = feature_cols
        dup_pred = dup_pred.reset_index(drop=True)
        dup_res = pd.concat([XY_dup_orig,X_dup,dup_pred], axis = 1)

        XY_del_orig = XY_del_orig.reset_index(drop=True)
        X_del = X_del.reset_index(drop=True)
        X_del.columns = feature_cols
        del_pred = del_pred.reset_index(drop=True)
        del_res = pd.concat([XY_del_orig,X_del,del_pred], axis = 1)

        # Create a prediction column
        def get_pred(row):
            pred_dict = {x:row[x] for x in ['BENIGN','VUS','PATHOGENIC']}
            pred_max = max(pred_dict.values())
            for k,v in pred_dict.items():
                if v == pred_max:
                    return k
        
        if len(del_res.index) != 0:
            del_res['PREDICTION'] = del_res.apply(get_pred, axis = 1)

        if len(dup_res.index) != 0:
            dup_res['PREDICTION'] = dup_res.apply(get_pred, axis = 1)
        
        bar_widgets[0] = progressbar.FormatLabel('Predictions complete')
        bar.update(4)


        return pd.concat([del_res, dup_res]).sort_values(['CHROMOSOME','START']).rename(columns={'gene_info':'INTERSECTING_GENES'})
    
    
    
        